import re

from ..config import settings
from ..config import config_interface
from .tools import cprint


def generate_part_number(category: str, part_pk: int) -> str:
    ''' Generate Internal Part Number (IPN) '''
    try:
        ipn = str(part_pk).zfill(settings.IPN_UNIQUE_ID_LENGTH)
    except:
        return None

    if settings.IPN_USE_FIXED_PREFIX:
        prefix_id = settings.IPN_PREFIX
    else:
        CATEGORY_CODES = config_interface.load_file(settings.CONFIG_CATEGORIES)['CODES']

        for key in CATEGORY_CODES.keys():
            if key in category:
                break
        try:
            prefix_id = CATEGORY_CODES[key]
        except:
            return None

    if prefix_id:
        ipn = '-'.join([prefix_id, ipn])

    if settings.IPN_USE_VARIANT_SUFFIX:
        ipn = '-'.join([ipn, settings.IPN_VARIANT_SUFFIX])

    return ipn


def compare(new_part_parameters: dict, db_part_parameters: dict, include_filters: list) -> bool:
    ''' Compare two InvenTree parts based on parameters (specs) '''
    try:
        for parameter, value in new_part_parameters.items():
            # Check for filters
            if include_filters:
                # Compare only parameters present in include_filters
                if parameter in include_filters and value != db_part_parameters[parameter]:
                    return False
            else:
                # Compare all parameters
                if value != db_part_parameters[parameter]:
                    return False
    except KeyError:
        cprint('[INFO]\tWarning: Failed to compare part with database', silent=settings.SILENT)
        return False

    return True


def clean_parameter_value(category: str, name: str, value: str) -> str:
    ''' Clean-up parameter value for consumption in InvenTree and KiCad '''
    category = category.lower()
    name = name.lower()

    # Parameter specific filters
    # Package
    if 'package' in name and 'size' not in name:
        space_split = value.split()

        # Return value before the space
        if len(space_split) > 1:
            value = space_split[0].replace(',', '')

    # Sizes
    if 'size' in name or \
            'height' in name or \
            'pitch' in name or \
            'outline' in name:
        # imperial = re.findall('[.0-9]*"', value)
        metric = re.findall('[.0-9]*mm', value)
        len_metric = len(metric)

        # Return only the metric dimensions
        if len_metric > 0 and len_metric <= 1:
            # One dimension
            if 'dia' in value.lower():
                # Check if diameter value
                value = '⌀' + metric[0]
            else:
                value = metric[0]
        elif len_metric > 1 and len_metric <= 2:
            # Two dimensions
            value = metric[0].replace('mm', '') + 'x' + metric[1]
        elif len_metric > 2 and len_metric <= 3:
            # Three dimensions
            value = metric[0].replace('mm', '') + 'x' + metric[1].replace('mm', '') + 'x' + metric[2]

    # Power
    if 'power' in name:
        # decimal = re.findall('[0-9]\.[0-9]*W', value)
        ratio = re.findall('[0-9]/[0-9]*W', value)

        # Return ratio
        if len(ratio) > 0:
            value = ratio[0]

    # ESR, DCR, RDS
    if 'esr' in name or \
            'dcr' in name or \
            'rds' in name:
        value = value.replace('Max', '').replace(' ', '').replace('Ohm', 'R')

    # Category specific filters
    # RESISTORS
    if 'resistor' in category:
        if 'resistance' in name:
            space_split = value.split()

            if len(space_split) > 1:
                resistance = space_split[0]
                unit = space_split[1]

                unit_filter = ['kOhms', 'MOhms', 'GOhms']
                if unit in unit_filter:
                    unit = unit.replace('Ohms', '').upper()
                else:
                    unit = unit.replace('Ohms', 'R')

                value = resistance + unit

    # General filters
    # Clean-up ranges
    separator = '~'
    if separator in value:
        split_space = value.split()
        first_value = split_space[0]
        second_value = split_space[2]

        # Substract digits, negative sign, points from first value to get unit
        unit = first_value.replace(re.findall('[-.0-9]*', first_value)[0], '')

        if unit:
            value = first_value.replace(unit, '') + separator + second_value

    # Remove parenthesis section
    if '(' in value:
        parenthesis = re.findall('\(.*\)', value)

        if parenthesis:
            for item in parenthesis:
                value = value.replace(item, '')

            # Remove leftover spaces
            value = value.replace(' ', '')

    # Remove spaces (for specific cases)
    if '@' in value:
        value = value.replace(' ', '')

    # Escape double-quote (else causes library error in KiCad)
    if '"' in value:
        value = value.replace('"', '\\"')

    # cprint(value)
    return value



import logging
import os
import digikey

from ..config import settings, config_interface

SEARCH_HEADERS = [
    'product_description',
    'detailed_description',
    'digi_key_part_number',
    'manufacturer',
    'manufacturer_part_number',
    'product_url',
    'primary_datasheet',
    'primary_photo',
]
PARAMETERS_MAP = [
    'parameters',
    'parameter',
    'value',
]


os.environ['DIGIKEY_STORAGE_PATH'] = settings.DIGIKEY_STORAGE_PATH
# Check if storage path exists, else create it
if not os.path.exists(os.environ['DIGIKEY_STORAGE_PATH']):
    os.makedirs(os.environ['DIGIKEY_STORAGE_PATH'], exist_ok=True)


def disable_api_logger():
    # Digi-Key API logger
    logging.getLogger('digikey.v3.api').setLevel(logging.WARNING)
    # Disable DEBUG
    logging.disable(logging.DEBUG)


def check_environment() -> bool:
    DIGIKEY_CLIENT_ID = os.environ.get('DIGIKEY_CLIENT_ID', None)
    DIGIKEY_CLIENT_SECRET = os.environ.get('DIGIKEY_CLIENT_SECRET', None)

    if not DIGIKEY_CLIENT_ID or not DIGIKEY_CLIENT_SECRET:
        return False

    return True


def setup_environment() -> bool:
    if not check_environment():
        # SETUP the Digikey authentication see https://developer.digikey.com/documentation/organization#production
        digikey_api_settings = config_interface.load_file(settings.CONFIG_DIGIKEY_API)
        os.environ['DIGIKEY_CLIENT_ID'] = digikey_api_settings['DIGIKEY_CLIENT_ID']
        os.environ['DIGIKEY_CLIENT_SECRET'] = digikey_api_settings['DIGIKEY_CLIENT_SECRET']

    return check_environment()


def get_default_search_keys():
    return [
        'product_description',
        'product_description',
        'revision',
        'keywords',
        'digi_key_part_number',
        'manufacturer',
        'manufacturer_part_number',
        'product_url',
        'primary_datasheet',
        'primary_photo',
    ]


def find_categories(part_details: str):
    ''' Find categories '''
    try:
        # print(part_details['limited_taxonomy']['children'][0]['value'])
        return part_details['limited_taxonomy']['children'][0]['value'], part_details['limited_taxonomy']['children'][0]['children'][0]['value']
    except:
        return None, None


def fetch_part_info(part_number: str) -> dict:
    ''' Fetch part data from API '''
    from ..wrapt_timeout_decorator import timeout

    part_info = {}
    if not setup_environment():
        return part_info

    @timeout(dec_timeout=20)
    def digikey_search_timeout():
        return digikey.product_details(part_number).to_dict()

    # Query part number
    try:
        part = digikey_search_timeout()
    except:
        part = None

    if not part:
        return part_info

    category, subcategory = find_categories(part)
    try:
        part_info['category'] = category
        part_info['subcategory'] = subcategory
    except:
        part_info['category'] = ''
        part_info['subcategory'] = ''

    headers = SEARCH_HEADERS

    for key in part:
        if key in headers:
            if key == 'manufacturer':
                part_info[key] = part['manufacturer']['value']
            else:
                part_info[key] = part[key]

    # Parameters
    part_info['parameters'] = {}
    [parameter_key, name_key, value_key] = PARAMETERS_MAP

    for parameter in range(len(part[parameter_key])):
        parameter_name = part[parameter_key][parameter][name_key]
        parameter_value = part[parameter_key][parameter][value_key]
        # Append to parameters dictionary
        part_info['parameters'][parameter_name] = parameter_value

    return part_info


def test_api_connect(check_content=False) -> bool:
    ''' Test method for API token '''
    setup_environment()

    test_success = True
    expected = {
        'product_description': 'RES 10K OHM 5% 1/16W 0402',
        'detailed_description': '10 kOhms ±5% 0.063W, 1/16W Chip Resistor 0402 (1005 '
                                'Metric) Automotive AEC-Q200 Thick Film',
        'digi_key_part_number': 'RMCF0402JT10K0CT-ND',
        'manufacturer': 'Stackpole Electronics Inc',
        'manufacturer_part_number': 'RMCF0402JT10K0',
        'product_url': 'https://www.digikey.com/product-detail/en/stackpole-electronics-inc/RMCF0402JT10K0/RMCF0402JT10K0CT-ND/1942936',
        'primary_datasheet': 'https://www.seielect.com/catalog/sei-rmcf_rmcp.pdf',
        'primary_photo': 'https://media.digikey.com/photos/Stackpole%20Photos/MFG_RMC%20SERIES.jpg',
    }

    test_part = fetch_part_info('RMCF0402JT10K0')

    # Check for response
    if not test_part:
        test_success = False
    
    if not check_content:
        return test_success
        
    # Check content of response
    if test_success:
        for key, value in expected.items():
            if test_part[key] != value:
                test_success = False
                break

    return test_success



from . import kicad_schlib

klib = kicad_schlib.ComponentLibManager()


def inventree_to_kicad(part_data: dict, library_path: str, template_path=None, show_progress=True) -> bool:
    ''' Create KiCad symbol from InvenTree part data '''
    return klib.add_component_to_library_from_inventree(component_data=part_data,
                                                        library_path=library_path,
                                                        template_path=template_path,
                                                        show_progress=show_progress)


def delete_part(part_number: str, library_path: str) -> bool:
    ''' Delete KiCad symbol from library '''
    return klib.delete_component_from_lib(part_number=part_number,
                                          library_path=library_path)



import copy
import os

from ..config import settings
from ..common import part_tools, progress
from ..common.tools import cprint
from ..config import config_interface
from ..database import inventree_api
from fuzzywuzzy import fuzz
from ..search import search_api, digikey_api, lcsc_api


def connect_to_server(timeout=5) -> bool:
    ''' Connect to InvenTree server using user settings '''
    connect = False
    settings.load_inventree_settings()

    try:
        connect = inventree_api.connect(server=settings.SERVER_ADDRESS,
                                        username=settings.USERNAME,
                                        password=settings.PASSWORD,
                                        connect_timeout=timeout)
    except TimeoutError:
        pass

    if not connect:
        if not settings.SERVER_ADDRESS:
            cprint('[TREE]\tError connecting to InvenTree server: missing server address')
            return connect
        if not settings.USERNAME:
            cprint('[TREE]\tError connecting to InvenTree server: missing username')
            return connect
        if not settings.PASSWORD:
            cprint('[TREE]\tError connecting to InvenTree server: missing password')
            return connect
        cprint('[TREE]\tError connecting to InvenTree server: invalid address, username or password')
    else:
        env = [env_type.name for env_type in settings.Environment
               if env_type.value == settings.environment][0]
        cprint(f'[TREE]\tSuccessfully connected to InvenTree server (ENV={env})', silent=settings.SILENT)

    return connect


def get_categories(part_info: dict, supplier_only=False) -> list:
    ''' Find categories from part supplier data, use "somewhat automatic" matching '''
    categories = [None, None]

    try:
        supplier_category = str(part_info['category'])
        supplier_subcategory = str(part_info['subcategory'])
    except KeyError:
        return categories

    # Return supplier category, if match not needed
    if supplier_only:
        categories[0] = supplier_category
        categories[1] = supplier_subcategory
        return categories

    function_filter = False
    # TODO: Make 'filter_parameter' user defined?
    filter_parameter = 'Function Type'

    # Check existing matches
    # Load inversed category map
    category_map = config_interface.load_supplier_categories_inversed(supplier_config_path=settings.CONFIG_DIGIKEY_CATEGORIES)

    try:
        for inventree_category in category_map.keys():
            for key, inventree_subcategory in category_map[inventree_category].items():
                if supplier_subcategory == key:
                    categories[0] = inventree_category
                    # Check if filtering by function
                    if inventree_subcategory.startswith(config_interface.FUNCTION_FILTER_KEY):
                        function_filter = True

                    # Save subcategory if not function filtered
                    if not function_filter:
                        categories[1] = inventree_subcategory

                    break
    except:
        pass

    # Function Filter
    if not categories[1] and function_filter:
        cprint(f'[INFO]\tSubcategory is filtered using "{filter_parameter}" parameter', silent=settings.SILENT, end='')
        # Load parameter map
        parameter_map = config_interface.load_category_parameters(categories[0], settings.CONFIG_SUPPLIER_PARAMETERS)
        # Build compare list
        compare = []
        for supplier_parameter, inventree_parameter in parameter_map.items():
            if (supplier_parameter in part_info['parameters'].keys() and inventree_parameter == filter_parameter):
                compare.append(part_info['parameters'][supplier_parameter])

        # Load subcategory map
        category_map = config_interface.load_supplier_categories(supplier_config_path=settings.CONFIG_DIGIKEY_CATEGORIES)[categories[0]]
        for inventree_subcategory in category_map.keys():
            for item in compare:
                fuzzy_match = fuzz.partial_ratio(inventree_subcategory, item)
                display_result = f'"{inventree_subcategory}" ?= "{item}"'.ljust(50)
                cprint(f'{display_result} => {fuzzy_match}', silent=settings.HIDE_DEBUG)
                if fuzzy_match >= settings.CATEGORY_MATCH_RATIO_LIMIT:
                    categories[1] = inventree_subcategory.replace(config_interface.FUNCTION_FILTER_KEY, '')
                    break

            if categories[1]:
                cprint('\t[ PASS ]', silent=settings.SILENT)
                break

    if not categories[1] and function_filter:
        cprint('\t[ FAILED ]', silent=settings.SILENT)

    # Automatic Match
    if not (categories[0] and categories[1]):
        # Load category map
        category_map = config_interface.load_supplier_categories(supplier_config_path=settings.CONFIG_DIGIKEY_CATEGORIES)

        def find_supplier_category_match(supplier_category: str, ignore_categories=False):
            # Check for match with Inventree categories
            category_match = None
            subcategory_match = None

            for inventree_category in category_map.keys():
                fuzzy_match = 0
                
                if not ignore_categories:
                    fuzzy_match = fuzz.partial_ratio(supplier_category, inventree_category)
                    display_result = f'"{supplier_category}" ?= "{inventree_category}"'.ljust(50)
                    cprint(f'{display_result} => {fuzzy_match}', silent=settings.HIDE_DEBUG)

                if fuzzy_match < settings.CATEGORY_MATCH_RATIO_LIMIT and category_map[inventree_category]:
                    # Compare to subcategories
                    for inventree_subcategory in category_map[inventree_category]:
                        fuzzy_match = fuzz.partial_ratio(supplier_category, inventree_subcategory)
                        display_result = f'"{supplier_category}" ?= "{inventree_subcategory}"'.ljust(50)
                        cprint(f'{display_result} => {fuzzy_match}', silent=settings.HIDE_DEBUG)

                        if fuzzy_match >= settings.CATEGORY_MATCH_RATIO_LIMIT:
                            subcategory_match = inventree_subcategory
                            break

                if fuzzy_match >= settings.CATEGORY_MATCH_RATIO_LIMIT:
                    category_match = inventree_category
                    break

            return category_match, subcategory_match

        # Find category and subcategories match
        category, subcategory = find_supplier_category_match(supplier_category)
        if category:
            categories[0] = category
        if subcategory:
            categories[1] = subcategory

        # Run match with supplier subcategory
        if not categories[0] or not categories[1]:
            if categories[0]:
                # If category was found: ignore them for the comparison
                category, subcategory = find_supplier_category_match(supplier_subcategory, ignore_categories=True)
            else:
                category, subcategory = find_supplier_category_match(supplier_subcategory)

        if category and not categories[0]:
            categories[0] = category
        if subcategory and not categories[1]:
            categories[1] = subcategory

    # Final checks
    if not categories[0]:
        cprint(f'[INFO]\tWarning: "{part_info["category"]}" did not match any supplier category ', silent=settings.SILENT)
    else:
        cprint(f'[INFO]\tCategory: "{categories[0]}"', silent=settings.SILENT)
    if not categories[1]:
        cprint(f'[INFO]\tWarning: "{part_info["subcategory"]}" did not match any supplier subcategory ', silent=settings.SILENT)
    else:
        cprint(f'[INFO]\tSubcategory: "{categories[1]}"', silent=settings.SILENT)

    return categories


def translate_supplier_to_inventree(supplier: str, part_info: dict, categories: list, skip_params=False) -> dict:
    ''' Using supplier part data and categories, fill-in InvenTree part dictionary '''

    def get_value_from_user_key(user_key: str, default_key: str, default_value=None) -> str:
        ''' Get value mapped from user search key, else default search key '''
        user_search_key = None
        if supplier == 'Digi-Key':
            user_search_key = settings.CONFIG_DIGIKEY.get(user_key, None)
        elif supplier == 'LCSC':
            user_search_key = settings.CONFIG_LCSC.get(user_key, None)
        
        # If no user key, use default
        if not user_search_key:
            return part_info.get(default_key, default_value)

        # Get value for user key, return value from default key if not found
        return part_info.get(user_search_key, part_info.get(default_key, default_value))

    # Copy template
    inventree_part = copy.deepcopy(settings.inventree_part_template)
    # Check that supplier argument is valid
    if not supplier or supplier not in settings.SUPPORTED_SUPPLIERS_API:
        return inventree_part
    # Get default keys
    if supplier == 'Digi-Key':
        default_search_keys = digikey_api.get_default_search_keys()
    elif supplier == 'LCSC':
        default_search_keys = lcsc_api.get_default_search_keys()

    # Insert data
    inventree_part["category"][0] = categories[0]
    inventree_part["category"][1] = categories[1]
    inventree_part['name'] = get_value_from_user_key('SEARCH_NAME', default_search_keys[0])
    inventree_part['description'] = get_value_from_user_key('SEARCH_DESCRIPTION', default_search_keys[1])
    # Revision
    inventree_part['revision'] = get_value_from_user_key('SEARCH_REVISION', default_search_keys[2], default_value=settings.INVENTREE_DEFAULT_REV)
    # Keywords (need to be after description)
    inventree_part['keywords'] = get_value_from_user_key('SEARCH_KEYWORDS', default_search_keys[1])
    inventree_part['supplier'] = {
        supplier: [
            get_value_from_user_key('SEARCH_SKU', default_search_keys[4]),
        ],
    }
    inventree_part['manufacturer'] = {
        get_value_from_user_key('SEARCH_MANUFACTURER', default_search_keys[5], default_value='manufacturer'): [
            get_value_from_user_key('SEARCH_MPN', default_search_keys[6], default_value='')
        ],
    }
    # Replace whitespaces in URL
    inventree_part['supplier_link'] = get_value_from_user_key('SEARCH_SUPPLIER_URL', default_search_keys[7], default_value='').replace(' ', '%20')
    inventree_part['datasheet'] = get_value_from_user_key('SEARCH_DATASHEET', default_search_keys[8], default_value='').replace(' ', '%20')
    # Image URL is not shown to user so force default key/value
    inventree_part['image'] = get_value_from_user_key('', default_search_keys[9], default_value='').replace(' ', '%20')

    # Load parameters map
    parameter_map = config_interface.load_category_parameters(category=inventree_part["category"][0],
                                                              supplier_config_path=settings.CONFIG_SUPPLIER_PARAMETERS, )

    if not skip_params:
        # Add Parameters
        if parameter_map:
            for supplier_param, inventree_param in parameter_map.items():
                # Some parameters may not be mapped
                if inventree_param not in inventree_part['parameters'].keys():
                    if supplier_param != 'Manufacturer Part Number':
                        try:
                            parameter_value = part_tools.clean_parameter_value(	category=categories[0],
                                                                                name=supplier_param,
                                                                                value=part_info['parameters'][supplier_param])
                            inventree_part['parameters'][inventree_param] = parameter_value
                        except:
                            cprint(f'[INFO]\tWarning: Parameter "{supplier_param}" not found in supplier data', silent=settings.SILENT)
                    else:
                        inventree_part['parameters'][inventree_param] = part_info['manufacturer_part_number']

            # Check for missing parameters and fill value with dash
            for inventree_param in parameter_map.values():
                if inventree_param not in inventree_part['parameters'].keys():
                    inventree_part['parameters'][inventree_param] = '-'
        else:
            cprint(f'[INFO]\tWarning: Parameter map for "{categories[0]}" does not exist or is empty', silent=settings.SILENT)

    return inventree_part


def translate_form_to_digikey(part_info: dict, categories: list, custom=False) -> dict:
    ''' Translate custom user part data and categories to Digi-Key API result format '''
    updated_part_info = {}

    updated_part_info['category'] = categories[0]
    updated_part_info['subcategory'] = categories[1]

    updated_part_info['product_name'] = part_info['name']
    updated_part_info['product_description'] = part_info['description']
    updated_part_info['revision'] = part_info['revision']
    updated_part_info['keywords'] = part_info['keywords']
    updated_part_info['digi_key_part_number'] = part_info['supplier_part_number']
    updated_part_info['manufacturer'] = part_info['manufacturer_name']
    updated_part_info['manufacturer_part_number'] = part_info['manufacturer_part_number']
    updated_part_info['primary_datasheet'] = part_info['datasheet']
    if custom:
        updated_part_info['primary_photo'] = ''

    return updated_part_info


def supplier_search(supplier: str, part_number: str, test_mode=False) -> dict:
    ''' Wrapper for supplier search, allow use of cached data (limited daily API calls) '''
    part_info = {}
    # Check part number exist
    if not part_number:
        cprint('\n[MAIN]\tError: Missing Part Number', silent=settings.SILENT)
        return part_info

    # Load from file if cache is enabled
    search_filename = settings.search_results['directory'] + supplier + '_' + part_number + settings.search_results['extension']

    # Get cached data
    part_info = search_api.load_from_file(search_filename, test_mode)
    if part_info:
        cprint(f'\n[MAIN]\tUsing {supplier} cached data for {part_number}', silent=settings.SILENT)
    else:
        cprint(f'\n[MAIN]\t{supplier} search for {part_number}', silent=settings.SILENT)
        if supplier == 'Digi-Key':
            part_info = digikey_api.fetch_part_info(part_number)
        elif supplier == 'LCSC':
            part_info = lcsc_api.fetch_part_info(part_number)

    # Check Digi-Key data exist
    if not part_info:
        cprint(f'[INFO]\tError: Failed to fetch data for "{part_number}"', silent=settings.SILENT)

    # Save search results
    if part_info:
        search_api.save_to_file(part_info, search_filename)

    return part_info


def inventree_create(supplier: str, part_info: dict, categories: list, kicad=False, symbol=None, footprint=None, show_progress=True, is_custom=False):
    ''' Create InvenTree part from supplier part data and categories '''
    # TODO: Make 'supplier' a variable for use with other APIs (eg. LCSC, Mouser, etc)
    part_pk = 0
    new_part = False

    # Translate to InvenTree part format
    inventree_part = translate_supplier_to_inventree(supplier=supplier,
                                                     part_info=part_info,
                                                     categories=categories,
                                                     skip_params=is_custom)

    if not inventree_part:
        cprint(f'\n[MAIN]\tError: Failed to process {supplier} data', silent=settings.SILENT)

    # Fetch category info from InvenTree part
    category_name = inventree_part['category'][0]
    subcategory_name = inventree_part['category'][1]
    category_pk = inventree_api.get_inventree_category_id(category_name)
    category_select = category_pk

    # Check if subcategory exists
    if subcategory_name:
        # Fetch subcategory id
        subcategory_pk = inventree_api.get_inventree_category_id(category_name=subcategory_name,
                                                                 parent_category_id=category_pk)
        if subcategory_pk > 0:
            category_select = subcategory_pk
        else:
            cprint(f'\n[TREE]\tWarning: Subcategory "{subcategory_name}" does not exist', silent=settings.SILENT)

    # Progress Update
    if show_progress and not progress.update_progress_bar_window():
        return new_part, part_pk, inventree_part

    if category_select <= 0:
        cprint(f'[ERROR]\tCategories ({category_name} - {subcategory_name}) does not exist in InvenTree', silent=settings.SILENT)
    else:
        # Check if part already exists
        part_pk = inventree_api.is_new_part(category_pk, inventree_part)
        # Part exists
        if part_pk > 0:
            cprint('[INFO]\tPart already exists, skipping.', silent=settings.SILENT)
            ipn = inventree_api.get_part_number(part_pk)
            # Update InvenTree part number
            inventree_part['IPN'] = ipn
            # Update InvenTree URL
            inventree_part['inventree_url'] = f'{settings.PART_URL_ROOT}{inventree_part["IPN"]}/'
        # Part is new
        else:
            new_part = True
            # Create a new Part
            # Use the pk (primary-key) of the category
            part_pk = inventree_api.create_part(category_id=category_select,
                                                name=inventree_part['name'],
                                                description=inventree_part['description'],
                                                revision=inventree_part['revision'],
                                                image=inventree_part['image'],
                                                keywords=inventree_part['keywords'])

            # Check part primary key
            if not part_pk:
                return new_part, part_pk, inventree_part
            # Progress Update
            if show_progress and not progress.update_progress_bar_window():
                return new_part, part_pk, inventree_part

            # Generate Internal Part Number
            cprint('\n[MAIN]\tGenerating Internal Part Number', silent=settings.SILENT)
            ipn = part_tools.generate_part_number(category_name, part_pk)
            cprint(f'[INFO]\tInternal Part Number = {ipn}', silent=settings.SILENT)
            # Update InvenTree part number
            ipn_update = inventree_api.set_part_number(part_pk, ipn)
            if not ipn_update:
                cprint('\n[INFO]\tError updating IPN', silent=settings.SILENT)
            inventree_part['IPN'] = ipn
            # Update InvenTree URL
            inventree_part['inventree_url'] = f'{settings.PART_URL_ROOT}{inventree_part["IPN"]}/'

    # Progress Update
    if show_progress and not progress.update_progress_bar_window():
        return new_part, part_pk, inventree_part

    if part_pk > 0:
        if new_part:
            cprint('[INFO]\tSuccess: Added new part to InvenTree', silent=settings.SILENT)
            if inventree_part['image']:
                # Add image
                image_result = inventree_api.upload_part_image(inventree_part['image'], part_pk)
                if not image_result:
                    cprint('[TREE]\tWarning: Failed to upload part image', silent=settings.SILENT)

        if kicad:
            # Create mandatory parameters (symbol & footprint)
            if symbol and ipn:
                kicad_symbol = symbol + ':' + ipn
            else:
                try:
                    kicad_symbol = settings.symbol_libraries_paths[category_name].split(os.sep)[-1].split('.')[0] + ':' + ipn
                except:
                    kicad_symbol = category_name.replace(' ', '_') + ':TBD'
            if footprint:
                kicad_footprint = footprint
            else:
                try:
                    supplier_package = inventree_part['parameters']['Package Type']
                    kicad_footprint = settings.footprint_lookup_table[category_name][supplier_package]
                except:
                    kicad_footprint = category_name.replace(' ', '_') + ':TBD'

            # Add symbol & footprint to InvenTree part
            inventree_part['parameters']['Symbol'] = kicad_symbol
            inventree_part['parameters']['Footprint'] = kicad_footprint

        if not inventree_part['parameters']:
            category_parameters = inventree_api.get_category_parameters(category_select)

            # Add category-defined parameters
            for parameter in category_parameters:
                inventree_part['parameters'][parameter[0]] = parameter[1]

        # Create parameters
        if len(inventree_part['parameters']) > 0:
            cprint('\n[MAIN]\tCreating parameters', silent=settings.SILENT)
            parameters_lists = [
                [],  # Store new parameters
                [],  # Store existings parameters
            ]
            for name, value in inventree_part['parameters'].items():
                parameter, is_new_parameter = inventree_api.create_parameter(part_id=part_pk, template_name=name, value=value)
                # Progress Update
                if show_progress and not progress.update_progress_bar_window(3):
                    return new_part, part_pk, inventree_part

                if is_new_parameter:
                    parameters_lists[0].append(name)
                else:
                    parameters_lists[1].append(name)

            if parameters_lists[0]:
                cprint('[INFO]\tSuccess: The following parameters were created:', silent=settings.SILENT)
                for item in parameters_lists[0]:
                    cprint(f'--->\t{item}', silent=settings.SILENT)
            if parameters_lists[1]:
                cprint('[TREE]\tWarning: The following parameters were skipped:', silent=settings.SILENT)
                for item in parameters_lists[1]:
                    cprint(f'--->\t{item}', silent=settings.SILENT)

        # Create manufacturer part
        manufacturer_name = None
        manufacturer_mpn = None
        # Extract manufacturer name and number from part data
        for key, values in inventree_part['manufacturer'].items():
            manufacturer_name = key
            manufacturer_mpn = values[0]
            break

        if manufacturer_mpn:
            cprint('\n[MAIN]\tCreating manufacturer part', silent=settings.SILENT)
            is_new_manufacturer_part = inventree_api.is_new_manufacturer_part(manufacturer_name=manufacturer_name,
                                                                              manufacturer_mpn=manufacturer_mpn)

            if not is_new_manufacturer_part:
                cprint('[INFO]\tManufacturer part already exists, skipping.', silent=settings.SILENT)
            else:
                # Create a new manufacturer part
                is_manufacturer_part_created = inventree_api.create_manufacturer_part(part_id=part_pk,
                                                                                      manufacturer_name=manufacturer_name,
                                                                                      manufacturer_mpn=manufacturer_mpn,
                                                                                      datasheet=inventree_part['datasheet'],
                                                                                      description=inventree_part['description'])

                if is_manufacturer_part_created:
                    cprint('[INFO]\tSuccess: Added new manufacturer part', silent=settings.SILENT)

        # Create supplier part
        try:
            supplier_sku = inventree_part['supplier'][supplier][0]
        except KeyError:
            supplier_sku = None

        if supplier_sku:
            cprint('\n[MAIN]\tCreating supplier part', silent=settings.SILENT)
            is_new_supplier_part = inventree_api.is_new_supplier_part(supplier_name=supplier,
                                                                      supplier_sku=supplier_sku)

            if not is_new_supplier_part:
                cprint('[INFO]\tSupplier part already exists, skipping.', silent=settings.SILENT)
            else:
                # Create a new supplier part
                is_supplier_part_created = inventree_api.create_supplier_part(part_id=part_pk,
                                                                              manufacturer_name=manufacturer_name,
                                                                              manufacturer_mpn=manufacturer_mpn,
                                                                              supplier_name=supplier,
                                                                              supplier_sku=inventree_part['supplier'][supplier],
                                                                              description=inventree_part['description'],
                                                                              link=inventree_part['supplier_link'])

                if is_supplier_part_created:
                    cprint('[INFO]\tSuccess: Added new supplier part', silent=settings.SILENT)

    # Progress Update
    if show_progress and not progress.update_progress_bar_window(3):
        pass

    return new_part, part_pk, inventree_part



import requests

SEARCH_HEADERS = [
    'productDescEn',
    'productIntroEn',
    'productCode',
    'brandNameEn',
    'productModel',
    'pdfUrl',
    'productImages',
]
PARAMETERS_MAP = [
    'paramVOList',
    'paramNameEn',
    'paramValueEn',
]


def get_default_search_keys():
    return [
        'productIntroEn',
        'productIntroEn',
        'revision',
        'keywords',
        'productCode',
        'brandNameEn',
        'productModel',
        '',
        'pdfUrl',
        'productImages',
    ]


def find_categories(part_details: str):
    ''' Find categories '''
    try:
        part_details['parentCatalogName']
        return part_details['parentCatalogName'], part_details['catalogName']
    except:
        return None, None


def fetch_part_info(part_number: str) -> dict:
    ''' Fetch part data from API '''
    from ..wrapt_timeout_decorator import timeout

    part_info = {}

    @timeout(dec_timeout=20)
    def search_timeout():
        url = 'https://wwwapi.lcsc.com/v1/products/detail?product_code=' + part_number
        response = requests.get(url)
        return response.json()

    # Query part number
    try:
        part = search_timeout()
    except:
        part = None

    if not part:
        return part_info

    category, subcategory = find_categories(part)
    try:
        part_info['category'] = category
        part_info['subcategory'] = subcategory
    except:
        part_info['category'] = ''
        part_info['subcategory'] = ''

    headers = SEARCH_HEADERS

    for key in part:
        if key in headers:
            if key == 'productImages':
                try:
                    part_info[key] = part['productImages'][0]
                except IndexError:
                    pass
            else:
                part_info[key] = part[key]

    # Parameters
    part_info['parameters'] = {}
    [parameter_key, name_key, value_key] = PARAMETERS_MAP

    for parameter in range(len(part[parameter_key])):
        parameter_name = part[parameter_key][parameter][name_key]
        parameter_value = part[parameter_key][parameter][value_key]
        # Append to parameters dictionary
        part_info['parameters'][parameter_name] = parameter_value

    return part_info


def test_api() -> bool:
    ''' Test method for API '''

    test_success = True
    expected = {
        'productIntroEn': 'C0G 25V ±5% 100pF 0201 Multilayer Ceramic Capacitors MLCC - SMD/SMT ROHS',
        'productCode': 'C2181718',
        'brandNameEn': 'TDK',
        'productModel': 'C0603C0G1E101J030BA',
    }

    test_part = fetch_part_info('C2181718')
        
    # Check content of response
    if test_success:
        for key, value in expected.items():
            if test_part[key] != value:
                print(f'"{test_part[key]}" <> "{value}"')
                test_success = False
                break

    return test_success



from .wrapt_timeout_decorator import timeout
from .wrap_helper import detect_unpickable_objects

# this needs to come after the module imports, otherwise circular import under windows
from . import __init__conf__
__title__ = __init__conf__.title
__version__ = __init__conf__.version
__name__ = __init__conf__.name
__url__ = __init__conf__.url
__author__ = __init__conf__.author
__author_email__ = __init__conf__.author_email
__shell_command__ = __init__conf__.shell_command



from ..config import settings
import validators
from ..common import part_tools
from ..common.tools import cprint, download_image
from ..config import config_interface

# Required to use local CA certificates on Linux
# For more details, refer to https://github.com/sparkmicro/Ki-nTree/pull/45
import platform
if platform.system() == 'Linux':
    import os
    cert_path = '/etc/ssl/certs/ca-certificates.crt'
    if os.path.isfile(cert_path):
        os.environ['REQUESTS_CA_BUNDLE'] = cert_path

# InvenTree
from inventree.api import InvenTreeAPI
from inventree.base import Parameter, ParameterTemplate
from inventree.company import Company, ManufacturerPart, SupplierPart
from inventree.part import Part, PartCategory


def connect(server: str, username: str, password: str, connect_timeout=5, silent=False) -> bool:
    ''' Connect to InvenTree server and create API object '''
    from ..wrapt_timeout_decorator import timeout
    global inventree_api

    @timeout(dec_timeout=connect_timeout)
    def get_inventree_api_timeout():
        return InvenTreeAPI(server, username=username, password=password)

    try:
        inventree_api = get_inventree_api_timeout()
    except:
        return False

    if inventree_api.token:
        return True
    return False


def get_inventree_category_id(category_name: str, parent_category_id=None) -> int:
    ''' Get InvenTree category ID from name, specificy parent if subcategory '''
    global inventree_api

    # Fetch all categories
    part_categories = PartCategory.list(inventree_api)

    for item in part_categories:
        if category_name == item.name:
            # Check parent id match (if passed as argument)
            match = True
            if parent_category_id:
                cprint(f'[TREE]\t{item.getParentCategory().pk} ?= {parent_category_id}', silent=settings.HIDE_DEBUG)
                if item.getParentCategory().pk != parent_category_id:
                    match = False
            if match:
                cprint(f'[TREE]\t{item.name} ?= {category_name} => True', silent=settings.HIDE_DEBUG)
                return item.pk
        else:
            cprint(f'[TREE]\t{item.name} ?= {category_name} => False', silent=settings.HIDE_DEBUG)

    return -1


def get_category_parameters(category_id: int) -> list:
    ''' Get all default parameter templates for category '''
    global inventree_api

    parameter_templates = []

    category = PartCategory(inventree_api, category_id)

    try:
        category_templates = category.get_category_parameter_templates(fetch_parent=True)
    except AttributeError:
        category_templates = None

    if category_templates:
        for template in category_templates:

            default_value = template.default_value
            if not default_value:
                default_value = '-'

            parameter_templates.append([template.parameter_template['name'], default_value])

    return parameter_templates


def get_part_number(part_id: int) -> str:
    ''' Get InvenTree part number from specified Part ID '''
    global inventree_api

    part = Part(inventree_api, part_id)
    return part.IPN


def set_part_number(part_id: int, ipn: str) -> bool:
    ''' Set InvenTree part number for specified Part ID '''
    global inventree_api

    part = Part(inventree_api, part_id)
    part._data['IPN'] = ipn
    part.save()

    if Part(inventree_api, part_id).IPN == ipn:
        return True
    else:
        return False


def is_new_part(category_id: int, part_info: dict) -> int:
    ''' Check if part exists based on parameters (or description) '''
    global inventree_api

    # Get category object
    part_category = PartCategory(inventree_api, category_id)

    # Fetch all parts from category and subcategories
    part_list = []
    part_list.extend(part_category.getParts())
    for subcategory in part_category.getChildCategories():
        part_list.extend(subcategory.getParts())

    # Extract parameter from part info
    new_part_parameters = part_info['parameters']

    template_list = ParameterTemplate.list(inventree_api)

    def fetch_template_name(template_id):
        for item in template_list:
            if item.pk == template_id:
                return item.name

    # Retrieve parent category name for parameters compare
    try:
        category_name = part_category.getParentCategory().name
    except AttributeError:
        category_name = part_category.name
    filters = config_interface.load_category_parameters_filters(category=category_name,
                                                                supplier_config_path=settings.CONFIG_PARAMETERS_FILTERS)
    # cprint(filters)

    for part in part_list:
        # Get part parameters
        db_part_parameters = part.getParameters()
        part_parameters = {}
        for parameter in db_part_parameters:
            parameter_name = fetch_template_name(parameter.template)
            parameter_value = parameter.data
            part_parameters[parameter_name] = parameter_value

        if new_part_parameters and part_parameters:
            # Compare database part with new part
            compare = part_tools.compare(new_part_parameters=new_part_parameters,
                                         db_part_parameters=part_parameters,
                                         include_filters=filters)
        else:
            # Compare with name and description
            compare = part_info['name'] == part.name and \
                part_info['description'] == part.description and \
                part_info['revision'] == part.revision

        if compare:
            cprint(f'\n[TREE]\tFound part match in database (pk = {part.pk})', silent=settings.HIDE_DEBUG)
            return part.pk

    cprint('\n[TREE]\tNo match found in database', silent=settings.HIDE_DEBUG)
    return 0


def create_category(parent: str, name: str):
    ''' Create InvenTree category, use parent for subcategories '''
    global inventree_api

    parent_id = 0
    is_new_category = False

    # Check if category already exists
    category_list = PartCategory.list(inventree_api)
    for category in category_list:
        if name == category.name:
            try:
                # Check if parents are the same
                if category.getParentCategory().name == parent:
                    # Return category ID
                    return category.pk, is_new_category
            except:
                return category.pk, is_new_category
        elif parent == category.name:
            # Get Parent ID
            parent_id = category.pk
        else:
            pass

    if parent:
        if parent_id > 0:
            category = PartCategory.create(inventree_api, {
                'name': name,
                'parent': parent_id,
            })

            is_new_category = True
        else:
            cprint(f'[TREE]\tError: Check parent category name ({parent})', silent=settings.SILENT)
            return -1, is_new_category
    else:
        # No parent
        category = PartCategory.create(inventree_api, {
            'name': name,
            'parent': None,
        })
        is_new_category = True

    return category.pk, is_new_category


def upload_part_image(image_url: str, part_id: int) -> bool:
    ''' Upload InvenTree part thumbnail'''
    global inventree_api

    # Get image full path
    image_name = f'{str(part_id)}_thumbnail.jpeg'
    image_location = settings.search_images + image_name

    # Download image (multiple attempts)
    if not download_image(image_url, image_location):
        return False

    # Upload image to InvenTree
    part = Part(inventree_api, part_id)
    if part:
        return part.uploadImage(image=image_location)
    else:
        return False


def create_part(category_id: int, name: str, description: str, revision: str, image: str, keywords=None) -> int:
    ''' Create InvenTree part '''
    global inventree_api

    part = Part.create(inventree_api, {
        'name': name,
        'description': description,
        'category': category_id,
        'keywords': keywords,
        'revision': revision,
        'active': True,
        'virtual': False,
        'component': True,
        'purchaseable': True,
    })

    if part:
        return part.pk
    else:
        return 0


def delete_part(part_id: int) -> bool:
    ''' Delete InvenTree part (only used for testing) '''
    global inventree_api

    part = Part(inventree_api, part_id)
    if part.pk:
        part._data['active'] = False
        # Remove image url (API rejects it as it is not a file)
        try:
            del part._data['image']
        except:
            pass
        part.save()
        return part.delete()
    else:
        return True


def create_company(company_name: str, manufacturer=False, supplier=False) -> bool:
    ''' Create InvenTree company '''
    global inventree_api

    if not manufacturer and not supplier:
        return None

    company = Company.create(inventree_api, {
        'name': company_name,
        'description': company_name,
        'is_customer': False,
        'is_supplier': supplier,
        'is_manufacturer': manufacturer,
    })

    return company


def get_company_id(company_name: str) -> int:
    ''' Get company (supplier/manufacturer) primary key (ID) '''
    global inventree_api

    company_list = Company.list(inventree_api)
    companies = {}
    for company in company_list:
        companies[company.name] = company.pk
    try:
        return companies[company_name]
    except:
        return 0


def is_new_manufacturer_part(manufacturer_name: str, manufacturer_mpn: str) -> bool:
    ''' Check if InvenTree manufacturer part exists to avoid duplicates '''
    global inventree_api

    # Fetch all companies
    cprint('[TREE]\tFetching manufacturers', silent=settings.HIDE_DEBUG)
    company_list = Company.list(inventree_api, is_manufacturer=True, is_customer=False)
    companies = {}
    for company in company_list:
        companies[company.name] = company

    try:
        # Get all parts
        part_list = companies[manufacturer_name].getManufacturedParts()
    except:
        part_list = None

    if part_list is None:
        # Create
        cprint(f'[TREE]\tCreating new manufacturer "{manufacturer_name}"', silent=settings.SILENT)
        create_company(
            company_name=manufacturer_name,
            manufacturer=True,
        )
        # Get all parts
        part_list = []

    for item in part_list:
        try:
            if manufacturer_mpn in item.MPN:
                cprint(f'[TREE]\t{item.MPN} ?= {manufacturer_mpn} => True', silent=settings.HIDE_DEBUG)
                return False
            else:
                cprint(f'[TREE]\t{item.MPN} ?= {manufacturer_mpn} => False', silent=settings.HIDE_DEBUG)
        except TypeError:
            cprint(f'[TREE]\t{item.MPN} ?= {manufacturer_mpn} => *** SKIPPED ***', silent=settings.HIDE_DEBUG)

    return True


def is_new_supplier_part(supplier_name: str, supplier_sku: str) -> bool:
    ''' Check if InvenTree supplier part exists to avoid duplicates '''
    global inventree_api

    # Fetch all companies
    cprint('[TREE]\tFetching suppliers', silent=settings.HIDE_DEBUG)
    company_list = Company.list(inventree_api, is_supplier=True, is_customer=False)
    companies = {}
    for company in company_list:
        companies[company.name] = company

    try:
        # Get all parts
        part_list = companies[supplier_name].getSuppliedParts()
    except:
        part_list = None

    if part_list is None:
        # Create
        cprint(f'[TREE]\tCreating new supplier "{supplier_name}"', silent=settings.SILENT)
        create_company(
            company_name=supplier_name,
            supplier=True,
        )
        # Get all parts
        part_list = []

    for item in part_list:
        if supplier_sku in item.SKU:
            cprint(f'[TREE]\t{item.SKU} ?= {supplier_sku} => True', silent=settings.HIDE_DEBUG)
            return False
        else:
            cprint(f'[TREE]\t{item.SKU} ?= {supplier_sku} => False', silent=settings.HIDE_DEBUG)

    return True


def create_manufacturer_part(part_id: int, manufacturer_name: str, manufacturer_mpn: str, description: str, datasheet: str) -> bool:
    ''' Create InvenTree manufacturer part

            part_id: Part the manufacturer data is linked to
            manufacturer: Company that manufactures the SupplierPart (leave blank if it is the sample as the Supplier!)
            MPN: Manufacture part number
            datasheet: Datasheet link
            description: Descriptive notes field
    '''
    global inventree_api

    # Get Manufacturer ID
    manufacturer_id = get_company_id(manufacturer_name)

    if manufacturer_id:
        # Validate datasheet link
        if not validators.url(datasheet):
            datasheet = ''

        manufacturer_part = ManufacturerPart.create(inventree_api, {
            'part': part_id,
            'manufacturer': manufacturer_id,
            'MPN': manufacturer_mpn,
            'link': datasheet,
            'description': description,
        })

        if manufacturer_part:
            return True
    else:
        cprint(f'[TREE]\tError: Manufacturer "{manufacturer_name}" not found (failed to create manufacturer part)',
               silent=settings.SILENT)

    return False


def create_supplier_part(part_id: int, manufacturer_name: str, manufacturer_mpn: str, supplier_name: str, supplier_sku: str, description: str, link: str) -> bool:
    ''' Create InvenTree supplier part

            part_id: Part the supplier data is linked to
            manufacturer_name: Manufacturer the supplier data is linked to
            manufacturer_mpn: MPN the supplier data is linked to
            supplier: Company that supplies this SupplierPart object
            SKU: Stock keeping unit (supplier part number)
            manufacturer: Company that manufactures the SupplierPart (leave blank if it is the sample as the Supplier!)
            MPN: Manufacture part number
            link: Link to part detail page on supplier's website
            description: Descriptive notes field
    '''
    global inventree_api

    # Get Supplier ID
    supplier_id = get_company_id(supplier_name)

    # Get Manufacturer ID
    manufacturer_id = get_company_id(manufacturer_name)
    if not manufacturer_id:
        # Unset MPN
        manufacturer_mpn = None

    if supplier_id:
        # Validate supplier link
        if not validators.url(link):
            link = ''

        supplier_part = SupplierPart.create(inventree_api, {
            'part': part_id,
            'manufacturer': manufacturer_id,
            'MPN': manufacturer_mpn,
            'supplier': supplier_id,
            'SKU': supplier_sku,
            'link': link,
            'description': description,
        })

        if supplier_part:
            return True
    else:
        cprint(f'[TREE]\tError: Supplier "{supplier_name}" not found (failed to create supplier part)',
               silent=settings.SILENT)

    return False


def create_parameter_template(name: str, units: str) -> int:
    ''' Create InvenTree parameter template '''
    global inventree_api

    parameter_templates = ParameterTemplate.list(inventree_api)
    for item in parameter_templates:
        if name == item.name:
            return 0

    parameter_template = ParameterTemplate.create(inventree_api, {
        'name': name,
        'units': units,
    })

    if parameter_template:
        return parameter_template.pk
    else:
        return 0


def create_parameter(part_id: int, template_name: int, value: str):
    ''' Create InvenTree part parameter based on template '''
    global inventree_api

    parameter_template_list = ParameterTemplate.list(inventree_api)

    template_id = 0
    for item in parameter_template_list:
        if template_name == item.name:
            template_id = item.pk
            break

    # Check if template_id already exists for this part
    part = Part(inventree_api, part_id)
    part_parameters = part.getParameters()
    is_new_part_parameters_template_id = True
    for item in part_parameters:
        # cprint(f'[TREE]\t{parameter.template} ?= {template_id}', silent=SILENT)
        if item.template == template_id:
            is_new_part_parameters_template_id = False
            break
    # cprint(part_parameters, silent=SILENT)

    '''
        Create parameter only if:
        - template exists
        - parameter does not exist for this part
    '''
    parameter = None
    if template_id > 0 and is_new_part_parameters_template_id:
        parameter = Parameter.create(inventree_api, {
            'part': part_id,
            'template': template_id,
            'data': value,
        })

    if parameter:
        return parameter.pk, is_new_part_parameters_template_id
    else:
        if template_id == 0:
            cprint(f'[TREE]\tError: Parameter template "{template_name}" does not exist', silent=settings.SILENT)
        return 0, False



import builtins
import json
import os
from shutil import copyfile


# CUSTOM PRINT METHOD
class pcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    ERROR = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

# Overload print function with custom pretty-print


def cprint(*args, **kwargs):
    # Check if silent is set
    try:
        silent = kwargs.pop('silent')
    except:
        silent = False
    if not silent:
        if type(args[0]) is dict:
            return builtins.print(json.dumps(*args, **kwargs, indent=4, sort_keys=True))
        else:
            try:
                args = list(args)
                if 'warning' in args[0].lower():
                    args[0] = f'{pcolors.WARNING}{args[0]}{pcolors.ENDC}'
                elif 'error' in args[0].lower():
                    args[0] = f'{pcolors.ERROR}{args[0]}{pcolors.ENDC}'
                elif 'fail' in args[0].lower():
                    args[0] = f'{pcolors.ERROR}{args[0]}{pcolors.ENDC}'
                elif 'success' in args[0].lower():
                    args[0] = f'{pcolors.OKGREEN}{args[0]}{pcolors.ENDC}'
                elif 'pass' in args[0].lower():
                    args[0] = f'{pcolors.OKGREEN}{args[0]}{pcolors.ENDC}'
                elif 'main' in args[0].lower():
                    args[0] = f'{pcolors.HEADER}{args[0]}{pcolors.ENDC}'
                elif 'skipping' in args[0].lower():
                    args[0] = f'{pcolors.BOLD}{args[0]}{pcolors.ENDC}'
                args = tuple(args)
            except:
                pass
            return builtins.print(*args, **kwargs, flush=True)
###


def create_library(library_path: str, symbol: str, template_lib: str):
    ''' Create library files if they don\'t exist '''

    if not os.path.exists(library_path):
        os.mkdir(library_path)
    new_lib_file = os.path.join(library_path, symbol + '.lib')
    new_dcm_file = new_lib_file.replace('.lib', '.dcm')
    template_dcm = template_lib.replace('.lib', '.dcm')
    if not os.path.exists(new_lib_file):
        copyfile(template_lib, new_lib_file)
    if not os.path.exists(new_dcm_file):
        copyfile(template_dcm, new_dcm_file)


def download_image(image_url: str, image_full_path: str, silent=False, retries=3) -> str:
    ''' Standard method to download image URL to local file '''
    import requests

    if not image_url:
        if not silent:
            cprint('[INFO]\tError: Missing image URL')
        return False

    def download(url):
        timeout = 3  # in seconds
        try:
            request = requests.get(url, timeout=timeout)
        except:
            if not silent:
                cprint(f'[INFO]\tWarning: Image download timed out ({timeout}s)')
            return None
        return request

    # Multiple download attempts
    for i in range(retries):
        request = download(image_url)
        if request:
            break

    # Still nothing
    if not request:
        return False

    with open(image_full_path, 'wb') as image:
        image.write(request.content)

    return True



''' DISCLAIMER
SchLib classes and methods were directly imported from: https://github.com/KiCad/kicad-library-utils
Class 'Component' init method was slightly modified to handle double-quote in part parameters:
https://github.com/KiCad/kicad-library-utils/pull/324
'''
# -*- coding: utf-8 -*-

import hashlib
import os.path
import shlex
import sys
from collections import OrderedDict


class Documentation(object):
    """
    A class to parse documentation files (dcm) of Schematic Libraries Files Format of the KiCad
    """
    line_keys = {
        'header': 'EESchema-DOCLIB',
        'start': '$CMP ',
        'description': 'D ',
        'keywords': 'K ',
        'datasheet': 'F ',
        'end': '$ENDCMP',
    }

    def __init__(self, filename, create=False):
        self.filename = filename
        self.components = OrderedDict()
        self.validFile = False
        self.header = None

        self.checksum = ""

        if create:
            if os.path.lexists(self.filename):
                sys.stderr.write("File already exists!\n")
                return
            else:
                self.validFile = True
                self.header = ["EESchema-DOCLIB  Version 2.0\n"]  # used for new dcm files

        else:
            if not os.path.isfile(self.filename):
                sys.stderr.write("DCM file '{filename}' does not exist\n".format(filename=self.filename))
                return
            else:
                self.validFile = True
                self.__parse()

    def __parse(self):
        f = open(self.filename, 'r')
        self.header = [f.readline()]

        if self.header and not self.line_keys['header'] in self.header[0]:
            self.header = None
            sys.stderr.write("'{fn}' is not a KiCad Documentation Library File\n".format(fn=self.filename))
            return False

        name = None
        f.seek(0)

        checksum_data = ''

        for line in f.readlines():
            checksum_data += line.strip()
            line = line.replace('\n', '')
            if line.startswith(Documentation.line_keys['start']):
                name = line[5:].strip()
                keywords = None
                description = None
                datasheet = None
            elif line.startswith(Documentation.line_keys['description']):
                description = line[2:]
            elif line.startswith(Documentation.line_keys['keywords']):
                keywords = line[2:]
            elif line.startswith(Documentation.line_keys['datasheet']):
                datasheet = line[2:]
            elif line.startswith(Documentation.line_keys['end']):
                self.components[name] = OrderedDict([('description', description), ('keywords', keywords), ('datasheet', datasheet)])
            # FIXME: we do not handle comments except separators around components
        f.close()

        try:
            md5 = hashlib.md5(checksum_data.encode('utf-8'))
        except UnicodeDecodeError:
            md5 = hashlib.md5(checksum_data)

        self.checksum = md5.hexdigest()

        return True

    def save(self, filename=None):
        if not self.validFile:
            return False

        if not filename:
            filename = self.filename

        to_write = self.header

        # Ensure that items are written in alphabetical order
        items = sorted(self.components.items(), key=lambda item: item[0])

        for name, doc in items:
            to_write.append('#\n')  # just spacer (no even in dcm format specification, but used everywhere)
            to_write.append(self.line_keys['start'] + name + '\n')
            for key in doc.keys():
                if(doc[key] is not None):
                    to_write.append(self.line_keys[key] + doc[key] + '\n')
            to_write.append(self.line_keys['end'] + '\n')
        to_write.append("#\n")  # again, spacer^^
        to_write.append("#End Doc Library\n")

        f = open(filename, 'w', newline='\n')
        f.writelines(to_write)
        f.close()

    def remove(self, name):
        if name in self.components.keys():  # delete only if it exists
            del self.components[name]

    def add(self, name, doc):
        if doc:  # do not create empty records
            self.components[name] = doc


class Component(object):
    """
    A class to parse components of Schematic Libraries Files Format of the KiCad
    """

    _DEF_KEYS = ['name', 'reference', 'unused', 'text_offset', 'draw_pinnumber', 'draw_pinname', 'unit_count', 'units_locked', 'option_flag']
    _F0_KEYS = ['reference', 'posx', 'posy', 'text_size', 'text_orient', 'visibility', 'htext_justify', 'vtext_justify']
    _FN_KEYS = ['name', 'posx', 'posy', 'text_size', 'text_orient', 'visibility', 'htext_justify', 'vtext_justify', 'fieldname']
    _ARC_KEYS = ['posx', 'posy', 'radius', 'start_angle', 'end_angle', 'unit', 'convert', 'thickness', 'fill', 'startx', 'starty', 'endx', 'endy']
    _CIRCLE_KEYS = ['posx', 'posy', 'radius', 'unit', 'convert', 'thickness', 'fill']
    _POLY_KEYS = ['point_count', 'unit', 'convert', 'thickness', 'points', 'fill']
    _RECT_KEYS = ['startx', 'starty', 'endx', 'endy', 'unit', 'convert', 'thickness', 'fill']
    _TEXT_KEYS = ['direction', 'posx', 'posy', 'text_size', 'text_type', 'unit', 'convert', 'text', 'italic', 'bold', 'hjustify', 'vjustify']
    _PIN_KEYS = ['name', 'num', 'posx', 'posy', 'length', 'direction', 'num_text_size', 'name_text_size', 'unit', 'convert', 'electrical_type', 'pin_type']

    _DRAW_KEYS = {'A': _ARC_KEYS, 'C': _CIRCLE_KEYS, 'P': _POLY_KEYS, 'S': _RECT_KEYS, 'T': _TEXT_KEYS, 'X': _PIN_KEYS}
    # _DRAW_ELEMS = {'arcs':'A', 'circles':'C', 'polylines':'P', 'rectangles':'S', 'texts':'T', 'pins':'X'}

    _KEYS = {'DEF': _DEF_KEYS, 'F0': _F0_KEYS, 'F': _FN_KEYS,
             'A': _ARC_KEYS, 'C': _CIRCLE_KEYS, 'P': _POLY_KEYS, 'S': _RECT_KEYS, 'T': _TEXT_KEYS, 'X': _PIN_KEYS}

    def __init__(self, data, comments, filename, documentation):
        self.comments = comments
        self.fplist = []
        self.aliases = OrderedDict()
        self.lib_filename = filename
        self.dcm_filename = documentation.filename
        building_fplist = False
        building_draw = False
        building_fields = False

        checksum_data = ''

        self.resetDraw()

        for line in data:
            checksum_data += line.strip()
            line = line.replace('\n', '')
            if '\\"' in line:
                import re
                line = re.findall(r'(?:[^\s,"]|"(?:\\.|[^"])*")+', line)
                # print(line)
            else:
                s = shlex.shlex(line)  # , posix=True)
                s.whitespace_split = True
                s.commenters = ''
                s.quotes = '"'
                line = list(s)

            if len(line) == 0:
                continue

            if line[0] in self._KEYS:
                key_list = self._KEYS[line[0]]
                values = line[1:] + ['' for n in range(len(key_list) - len(line[1:]))]

            if line[0] == 'DEF':
                building_fields = True
                self.definition = dict(zip(self._DEF_KEYS, values))

            elif line[0] == 'ALIAS':
                for alias in line[1:]:
                    self.aliases[alias] = self.getDocumentation(documentation, alias)

            elif line[0] == '$FPLIST':
                building_fields = False
                building_fplist = True
                self.fplist = []

            elif line[0] == '$ENDFPLIST':
                building_fplist = False

            elif line[0] == 'DRAW':
                building_draw = True
                self.resetDraw()
                self.drawOrdered = []  # list of draw elements references, needed to preserve line ordering

            elif line[0] == 'ENDDRAW':
                building_draw = False

            else:
                if building_fplist:
                    self.fplist.append(line[0])

                elif building_draw:
                    if line[0] == 'A':
                        self.draw['arcs'].append(dict(zip(self._ARC_KEYS, values)))
                        self.drawOrdered.append(['A', self.draw['arcs'][-1]])
                    if line[0] == 'C':
                        self.draw['circles'].append(dict(zip(self._CIRCLE_KEYS, values)))
                        self.drawOrdered.append(['C', self.draw['circles'][-1]])
                    if line[0] == 'P':  # mixing X an Y points into 1 list in not handy
                        n_points = int(line[1])
                        points = line[5:5 + (2 * n_points)]
                        values = line[1:5] + [points]
                        if len(line) > (5 + len(points)):
                            values += [line[-1]]
                        else:
                            values += ['']
                        self.draw['polylines'].append(dict(zip(self._POLY_KEYS, values)))
                        self.drawOrdered.append(['P', self.draw['polylines'][-1]])
                    if line[0] == 'S':
                        self.draw['rectangles'].append(dict(zip(self._RECT_KEYS, values)))
                        self.drawOrdered.append(['S', self.draw['rectangles'][-1]])
                    if line[0] == 'T':
                        self.draw['texts'].append(dict(zip(self._TEXT_KEYS, values)))
                        self.drawOrdered.append(['T', self.draw['texts'][-1]])
                    if line[0] == 'X':
                        self.draw['pins'].append(dict(zip(self._PIN_KEYS, values)))
                        self.drawOrdered.append(['X', self.draw['pins'][-1]])

                elif building_fields:
                    if line[0] == 'F0':
                        self.fields = []
                        self.fields.append(dict(zip(self._F0_KEYS, values)))

                    elif line[0][0] == 'F':
                        values = line[1:] + ['' for n in range(len(self._FN_KEYS) - len(line[1:]))]
                        self.fields.append(dict(zip(self._FN_KEYS, values)))

        # perform checksum calculation
        try:
            md5 = hashlib.md5(checksum_data.encode('utf-8'))
        except UnicodeDecodeError:
            md5 = hashlib.md5(checksum_data)
        self.checksum = md5.hexdigest()

        # define some shortcuts
        self.name = self.definition['name']
        self.reference = self.definition['reference']
        self.pins = self.draw['pins']

        # get documentation
        self.documentation = self.getDocumentation(documentation, self.name)

    def resetDraw(self):
        self.draw = {
            'arcs': [],
            'circles': [],
            'polylines': [],
            'rectangles': [],
            'texts': [],
            'pins': []
        }

    def getDocumentation(self, documentation, name):
        try:
            if name.startswith('~'):
                return documentation.components[name[1:(len(name))]]
            else:
                return documentation.components[name]
        except KeyError:
            return {}

    def getPinsByName(self, name):
        pins = []
        for pin in self.pins:
            if pin['name'] == name:
                pins.append(pin)

        return pins

    def getPinByNumber(self, num):
        for pin in self.draw['pins']:
            if pin['num'] == str(num):
                return pin

        return None

    def filterPins(self, name=None, direction=None, electrical_type=None):
        pins = []

        for pin in self.pins:
            if ((name and pin['name'] == name) or (direction and pin['direction'] == direction) or (electrical_type and pin['electrical_type'] == electrical_type)):
                pins.append(pin)

        return pins

    def isNonBOMSymbol(self):
        return self.reference.startswith('#')

    def isPowerSymbol(self):
        return self.definition['option_flag'] == 'P'

    def isPossiblyPowerSymbol(self):
        return (self.reference == '#PWR')

    def isGraphicSymbol(self):
        return self.isNonBOMSymbol() and len(self.pins) == 0

    # heuristics, which tries to determine whether this is a "small" component (resistor, capacitor, LED, diode, transistor, ...)
    def isSmallComponentHeuristics(self):
        if len(self.pins) <= 2:
            return True

        # If there is only a single filled rectangle, we assume that it is the
        # main symbol outline.
        drawing = self.draw
        filled_rects = [rect for rect in drawing['rectangles']
                        if rect['fill'] == 'f']

        # if there is no filled rectangle as symbol outline and we have 3 or 4 pins, we assume this
        # is a small symbol
        if len(self.pins) >= 3 and len(self.pins) <= 4 and len(filled_rects) == 0:
            return True

        return False


class SchLib(object):
    """
    A class to parse Schematic Libraries Files Format of the KiCad
    """

    line_keys = {
        'header': 'EESchema-LIBRARY',
    }

    def __init__(self, filename, create=False):
        self.filename = filename
        self.header = None
        self.components = []
        self.validFile = False

        self.checksum = ""

        self.documentation = Documentation(self.libToDcmFilename(self.filename), create)

        if create:
            if os.path.lexists(self.filename):
                sys.stderr.write("File already exists!\n")
                return
            else:
                self.validFile = True
                self.header = ['EESchema-LIBRARY Version 2.3\n', '#encoding utf-8\n']

        else:
            if not os.path.isfile(self.filename):
                sys.stderr.write("Library file '{filename}' does not exist\n".format(filename=self.filename))
                return
            else:
                self.validFile = True
                self.__parse()

    def libToDcmFilename(self, filename):
        dir_path = os.path.dirname(os.path.realpath(filename))
        filename = os.path.splitext(os.path.basename(filename))
        return os.path.join(dir_path, filename[0] + '.dcm')

    def __parse(self):
        f = open(self.filename, 'r')

        checksum_data = ""

        self.header = [f.readline()]

        checksum_data += self.header[0]

        if self.header and not SchLib.line_keys['header'] in self.header[0]:
            sys.stderr.write("'{fn}' is not a KiCad Schematic Library File\n".format(fn=self.filename))
            return False

        self.header.append(f.readline())
        building_component = False

        comments = []
        for line in f.readlines():

            checksum_data += line.strip()

            if line.startswith('#'):
                comments.append(line)

            elif line.startswith('DEF'):
                building_component = True
                component_data = []
                component_data.append(line)

            elif building_component:
                component_data.append(line)
                if line.startswith('ENDDEF'):
                    building_component = False
                    self.components.append(Component(component_data, comments, self.filename, self.documentation))
                    comments = []
        f.close()

        # perform checksum calculation
        try:
            md5 = hashlib.md5(checksum_data.encode('utf-8'))
        except UnicodeDecodeError:
            md5 = hashlib.md5(checksum_data)
        self.checksum = md5.hexdigest()

        return True

    def validChecksum(self):
        if len(self.checksum) == 0:
            return False
        if len(self.documentation.checksum) == 0:
            return False

        return True

    def compareChecksum(self, otherlib):

        if not self.validChecksum() or not otherlib.validChecksum():
            return False

        return self.checksum == otherlib.checksum and self.documentation.checksum == otherlib.documentation.checksum

    def getComponentByName(self, name):
        for component in self.components:
            if component.definition['name'] == name:
                return component

        return None

    def getComponentCount(self, unique=False):
        count = 0

        for cmp in self.components:
            count += 1

            if unique:
                continue

            count += len(cmp.aliases)

        return count

    def removeComponent(self, name):
        component = self.getComponentByName(name)
        for alias in component.aliases.keys():
            self.documentation.remove(alias)
        self.documentation.remove(name)
        self.components.remove(component)
        return component

    def addComponent(self, component):
        if component not in self.components:
            self.components.append(component)
            self.documentation.add(component.name, component.documentation)
            for alias in component.aliases.keys():
                self.documentation.add(alias, component.aliases[alias])

    def save(self, filename=None):
        if not self.validFile:
            return False

        if not filename:
            filename = self.filename

        self.documentation.save(self.libToDcmFilename(filename))

        # insert the header
        to_write = self.header

        # Ensure that the components are sorted by name!
        components = sorted(self.components, key=lambda cmp: cmp.name)

        # insert the components
        for component in components:
            # append the component comments
            to_write += component.comments

            # DEF
            line = 'DEF '
            for key in Component._DEF_KEYS:
                line += component.definition[key] + ' '

            line = line.rstrip() + '\n'
            to_write.append(line)

            # FIELDS
            line = 'F'
            for i, f in enumerate(component.fields):
                line = "F{n} ".format(n=i)

                if i == 0:
                    keys_list = Component._F0_KEYS
                else:
                    keys_list = Component._FN_KEYS

                for k, key in enumerate(keys_list):
                    key_val = component.fields[i][key]

                    if k == 0 and not key_val.startswith('"'):
                        key_val = '"' + key_val + '"'

                    line += key_val + ' '

                line = line.rstrip() + '\n'
                to_write.append(line)

            # ALIAS
            if len(component.aliases) > 0:
                line = 'ALIAS '
                for alias in component.aliases.keys():
                    line += alias + ' '

                line = line.rstrip() + '\n'
                to_write.append(line)

            # $FPLIST
            if len(component.fplist) > 0:
                to_write.append('$FPLIST\n')
                for fp in component.fplist:
                    to_write.append(' ' + fp + '\n')

            # $ENDFPLIST
                to_write.append('$ENDFPLIST\n')

            # DRAW
            to_write.append('DRAW\n')
            for elem in component.drawOrdered:
                item = elem[1]
                keys_list = Component._DRAW_KEYS[elem[0]]  # 'A' -> keys of all properties of arc
                line = elem[0] + ' '  # 'arcs' -> 'A'
                for k in keys_list:
                    if k == 'points':
                        for i in item['points']:
                            line += '{0} '.format(i)
                    else:
                        line += item[k] + ' '

                line = line.rstrip() + '\n'
                to_write.append(line)

            # ENDDRAW
            to_write.append('ENDDRAW\n')

            # ENDDEF
            to_write.append('ENDDEF\n')

        # insert the footer
        to_write.append('#\n')
        to_write.append('#End Library\n')

        f = open(filename, 'w', newline='\n')
        f.writelines(to_write)
        f.close()



import base64
import copy
import os

import yaml
from ..common.tools import cprint

FUNCTION_FILTER_KEY = '__'


def load_file(file_path: str, silent=True) -> dict:
    ''' Safe load YAML file '''
    try:
        with open(file_path, 'r') as file:
            try:
                data = yaml.safe_load(file)
            except yaml.YAMLError as exc:
                print(exc)
                return None
    except FileNotFoundError:
        if not silent:
            cprint(f'[ERROR] File {file_path} does not exists!')
        return None

    return data


def dump_file(data: dict, file_path: str) -> bool:
    ''' Safe dump YAML file '''
    with open(file_path, 'w') as file:
        try:
            yaml.safe_dump(data, file, default_flow_style=False)
        except yaml.YAMLError as exc:
            print(exc)
            return False

    return True


def load_user_paths(home_dir='') -> dict:
    ''' Load user config and cache paths '''

    user_settings_file = os.path.join(home_dir, 'settings.yaml')
    user_config = load_file(user_settings_file)

    if not user_config:
        user_config = {
            'USER_FILES': os.path.join(home_dir, 'user', ''),
            'USER_CACHE': os.path.join(home_dir, 'cache', ''),
        }
        dump_file(user_config, user_settings_file)

    return user_config


def load_user_config_files(path_to_root: str, path_to_user_files: str, silent=True) -> bool:
    ''' Load InvenTree user configuration '''
    result = True

    def load_config(path):
        for template_file in os.listdir(path):
            filename = os.path.basename(template_file)
            template_data = load_file(path + template_file)
            try:
                user_data = load_file(path_to_user_files + filename)
                # Join user data to template data
                user_settings = {**template_data, **user_data}
            except:
                user_settings = template_data

            dump_file(user_settings, path_to_user_files + filename)

    # Load User settings
    try:
        config_files = os.path.join(path_to_root, 'settings', '')
        load_config(config_files)
    except:
        cprint('[INFO]\tWarning: Failed to load User settings', silent=silent)
        result = False
    # Load Search API configuration files
    try:
        config_files = os.path.join(path_to_root, 'search', '')
        load_config(config_files)
    except:
        cprint('[INFO]\tWarning: Failed to load Search API configuration', silent=silent)
        result = False
    # Load Digi-Key configuration files
    try:
        config_files = os.path.join(path_to_root, 'digikey', '')
        load_config(config_files)
    except:
        cprint('[INFO]\tWarning: Failed to load Digi-Key configuration', silent=silent)
        result = False
    # Load LCSC configuration files
    try:
        config_files = os.path.join(path_to_root, 'lcsc', '')
        load_config(config_files)
    except:
        cprint('[INFO]\tWarning: Failed to load LCSC configuration', silent=silent)
        result = False
    # Load InvenTree configuration files
    try:
        config_files = os.path.join(path_to_root, 'inventree', '')
        load_config(config_files)
    except:
        cprint('[INFO]\tWarning: Failed to load InvenTree configuration', silent=silent)
        result = False
    # Load KiCad configuration files
    try:
        config_files = os.path.join(path_to_root, 'kicad', '')
        load_config(config_files)
    except:
        cprint('[INFO]\tWarning: Failed to load KiCad configuration', silent=silent)
        result = False

    return result


def load_inventree_user_settings(user_config_path: str) -> dict:
    ''' Load InvenTree user settings from file '''
    user_settings = load_file(user_config_path)

    try:
        password = user_settings.get('PASSWORD', None)
    except AttributeError:
        return user_settings

    try:
        # Use base64 encoding to make password unreadable inside the file
        user_settings['PASSWORD'] = base64.b64decode(password).decode()
    except TypeError:
        user_settings['PASSWORD'] = ''

    return user_settings


def save_inventree_user_settings(enable: bool, server: str, username: str, password: str, user_config_path: str):
    ''' Save InvenTree user settings to file '''
    user_settings = {}

    user_settings['ENABLE'] = enable
    user_settings['SERVER_ADDRESS'] = server
    user_settings['USERNAME'] = username
    # Use base64 encoding to make password unreadable inside the file
    user_settings['PASSWORD'] = base64.b64encode(password.encode())

    return dump_file(user_settings, user_config_path)


def load_library_path(user_config_path: str, silent=False):
    ''' Load KiCad library from KiCad settings file '''
    user_settings = load_file(user_config_path)

    try:
        if not user_settings['KICAD_SYMBOLS_PATH'] and not silent:
            print('[INFO]\tEmpty KiCad library path')
        return user_settings['KICAD_SYMBOLS_PATH']
    except:
        # If not defined: use application root folder
        return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def add_library_path(user_config_path: str, category: str, symbol_library: str) -> bool:
    ''' Save KiCad library to KiCad settings file '''
    user_settings = load_file(user_config_path)

    if category:
        index = category
    else:
        index = symbol_library

    if not user_settings['KICAD_LIBRARIES']:
        user_settings['KICAD_LIBRARIES'] = {}

    try:
        if symbol_library not in user_settings['KICAD_LIBRARIES'][index]:
            user_settings['KICAD_LIBRARIES'][index].append(symbol_library)
    except:
        user_settings['KICAD_LIBRARIES'][index] = [symbol_library]

    return dump_file(user_settings, user_config_path)


def load_libraries_paths(user_config_path: str, library_path: str) -> dict:
    ''' Construct KiCad library files names and paths from KiCad settings file '''
    user_settings = load_file(user_config_path)

    if not os.path.exists(library_path):
        return None

    found_library_files = []
    for file in os.listdir(library_path):
        if file.endswith('.lib'):
            found_library_files.append(file.replace('.lib', ''))

    symbol_libraries_paths = {}
    assigned_files = []
    try:
        for category, libraries in user_settings['KICAD_LIBRARIES'].items():
            symbol_libraries_paths[category] = {}
            if libraries:
                for library in libraries:
                    if library in found_library_files:
                        symbol_libraries_paths[category][library] = library_path + \
                            library + '.lib'
                        assigned_files.append(library)
    except:
        pass

    for file in found_library_files:
        if file not in assigned_files:
            try:
                symbol_libraries_paths['uncategorized'].append(file)
            except:
                symbol_libraries_paths['uncategorized'] = [file]
    try:
        symbol_libraries_paths['uncategorized'] = sorted(
            symbol_libraries_paths['uncategorized'])
    except:
        pass

    # Check that library paths are loaded
    path_loaded = False
    for category, paths in symbol_libraries_paths.items():
        if paths:
            path_loaded = True
            break
    if not path_loaded:
        return None

    # print(symbol_libraries_paths)
    return symbol_libraries_paths


def load_templates_paths(user_config_path: str, template_path: str) -> dict:
    ''' Construct KiCad template files names and paths from KiCad settings file '''
    symbol_templates_paths = {}
    if not template_path:
        return symbol_templates_paths

    # Load configuration file
    user_settings = load_file(user_config_path)

    try:
        for category in user_settings['KICAD_TEMPLATES'].keys():
            for subcategory, file_name in user_settings['KICAD_TEMPLATES'][category].items():
                if subcategory == 'Default' and not file_name:
                    file_name = 'default'
                if file_name:
                    try:
                        symbol_templates_paths[category][subcategory] = template_path + \
                            file_name + '.lib'
                    except:
                        symbol_templates_paths[category] = {
                            subcategory: template_path + file_name + '.lib'}
    except:
        pass

    return symbol_templates_paths


def load_footprint_paths(user_config_path: str, footprint_path: str) -> dict:
    ''' Construct KiCad footprint folder names and paths from KiCad settings file '''
    user_settings = load_file(user_config_path)

    if not os.path.exists(footprint_path):
        return None

    found_library_folders = [	item.replace('.pretty', '') for item in os.listdir(footprint_path)
                              if os.path.isdir(footprint_path + item)]

    footprint_libraries_paths = {}
    assigned_folders = []
    try:
        for category, libraries in user_settings['KICAD_FOOTPRINTS'].items():
            footprint_libraries_paths[category] = {}
            if libraries:
                for folder in libraries:
                    footprint_libraries_paths[category][folder] = footprint_path + \
                        folder + '.pretty'
                    assigned_folders.append(folder)
    except:
        pass

    for folder in found_library_folders:
        if folder not in assigned_folders:
            try:
                footprint_libraries_paths['uncategorized'].append(folder)
            except:
                footprint_libraries_paths['uncategorized'] = [folder]

        # Sort uncategorized library paths
        footprint_libraries_paths['uncategorized'] = sorted(
            footprint_libraries_paths.get('uncategorized', []))

    return footprint_libraries_paths


def add_footprint_library(user_config_path: str, category: str, library_folder: str) -> bool:
    ''' Add KiCad footprint folder name to KiCad settings file '''
    user_settings = load_file(user_config_path)

    if category:
        index = category
    else:
        index = library_folder

    if not user_settings['KICAD_FOOTPRINTS']:
        user_settings['KICAD_FOOTPRINTS'] = {}

    try:
        if library_folder not in user_settings['KICAD_FOOTPRINTS'][index]:
            user_settings['KICAD_FOOTPRINTS'][index].append(library_folder)
    except:
        user_settings['KICAD_FOOTPRINTS'][index] = [library_folder]

    return dump_file(user_settings, user_config_path)


def load_supplier_categories(supplier_config_path: str, clean=False) -> dict:
    ''' Load Supplier category mapping from Supplier settings file '''
    supplier_categories = load_file(supplier_config_path)

    if clean:
        clean_supplier_categories = copy.deepcopy(supplier_categories)

        for category in supplier_categories:
            for subcategory in supplier_categories[category]:
                if FUNCTION_FILTER_KEY in subcategory:
                    clean_supplier_categories[category][subcategory.replace(FUNCTION_FILTER_KEY, '')] \
                        = supplier_categories[category][subcategory]
                    del clean_supplier_categories[category][subcategory]

        return clean_supplier_categories

    # print(supplier_categories)
    return supplier_categories


def load_supplier_categories_inversed(supplier_config_path: str) -> dict:
    ''' Load Supplier category mapping from Supplier settings file (inversed relation) '''
    supplier_categories = load_file(supplier_config_path)

    try:
        supplier_categories_inversed = {}
        for category in supplier_categories.keys():
            if supplier_categories[category]:
                for user, supplier in supplier_categories[category].items():
                    # Supplier is list type
                    if supplier:
                        if category not in supplier_categories_inversed.keys():
                            supplier_categories_inversed[category] = {}
                        for item in supplier:
                            supplier_categories_inversed[category][item] = user
    except:
        return None

    # print(supplier_categories_inversed)
    return supplier_categories_inversed


def sync_inventree_supplier_categories(inventree_config_path: str, supplier_config_path: str) -> dict:
    ''' Synchronize supplier categories dict from InvenTree categories '''
    inventree_categories = load_file(inventree_config_path)['CATEGORIES']
    supplier_categories = load_supplier_categories(supplier_config_path, clean=True)
    updated_supplier_categories = copy.deepcopy(supplier_categories)

    try:
        for category in inventree_categories:
            if category not in supplier_categories.keys():
                updated_supplier_categories[category] = inventree_categories[category]
    except:
        pass

    return updated_supplier_categories


def add_supplier_category(categories: dict, supplier_config_path: str) -> bool:
    ''' Add Supplier category mapping to Supplier settings file

                    categories = {
                                    'Capacitors':
                                                    { 'Tantalum': 'Tantalum Capacitors' }
                    }
    '''
    try:
        supplier_categories = load_file(supplier_config_path)
    except:
        return None

    for category in categories.keys():
        for user_subcategory, supplier_category in categories[category].items():
            try:
                supplier_category_keys = supplier_categories[category].keys()
            except:
                supplier_categories[category] = {
                    user_subcategory: [supplier_category]}
                break

            # Function filtered
            inventree_subcategory_filter = FUNCTION_FILTER_KEY + user_subcategory
            if inventree_subcategory_filter in supplier_category_keys:
                try:
                    if supplier_category not in supplier_categories[category][inventree_subcategory_filter]:
                        supplier_categories[category][inventree_subcategory_filter].append(
                            supplier_category)
                    break
                except:
                    pass

                try:
                    supplier_categories[category][inventree_subcategory_filter] = [
                        supplier_category]
                    break
                except:
                    pass
            else:
                try:
                    if supplier_category not in supplier_categories[category][user_subcategory]:
                        supplier_categories[category][user_subcategory].append(
                            supplier_category)
                    break
                except:
                    pass

                try:
                    supplier_categories[category][user_subcategory] = [
                        supplier_category]
                    break
                except:
                    pass

            return False

    return dump_file(supplier_categories, supplier_config_path)


def load_category_parameters(category: str, supplier_config_path: str) -> dict:
    ''' Load Supplier parameters mapping from Supplier settings file '''
    try:
        category_parameters = load_file(supplier_config_path)[category]
    except:
        return None

    category_parameters_inversed = {}
    for parameter in category_parameters.keys():
        if category_parameters[parameter]:
            for supplier_parameter in category_parameters[parameter]:
                category_parameters_inversed[supplier_parameter] = parameter

    # print(category_parameters_inversed)
    return category_parameters_inversed


def load_category_parameters_filters(category: str, supplier_config_path: str) -> list:
    ''' Load Supplier parameters filters from Supplier settings file '''
    try:
        parameters_filters = load_file(supplier_config_path)[category]
    except:
        return []

    # print(parameters_filters)
    return parameters_filters



import os

from ..config import settings
from ..common import progress
from ..common.tools import cprint
from schlib import SchLib


# KiCad Component Library Manager
class ComponentLibManager(object):
    def __init__(self):
        pass

    def is_component_in_library(self, library_instance, component_part_number):
        ''' Check if component already exists in library '''
        for component in library_instance.components:
            cprint(f'[DBUG]\t{component.definition["name"]} ?= {component_part_number}', silent=settings.HIDE_DEBUG)
            if component.definition['name'] == component_part_number:
                cprint(f'[KCAD]\tWarning: Component {component_part_number} already in library', silent=settings.SILENT)
                return True

        return False

    def add_component_to_library_from_inventree(self, component_data, library_path, template_path=None, show_progress=True):
        ''' Create component (symbol) in KiCad library '''
        part_in_lib = False
        new_part = False
        category = component_data['category'][0]
        subcategory = component_data['category'][1]

        # Load library and template paths
        cprint(f'[KCAD]\tlibrary_path: {library_path}', silent=settings.SILENT)

        if not template_path:
            # Fetch template path
            try:
                template_path = settings.symbol_templates_paths[category][subcategory]
            except:
                template_path = settings.symbol_templates_paths[category]['Default']

        # Check files exist
        if not os.path.isfile(library_path):
            cprint(f'[KCAD]\tError loading library file ({library_path})', silent=settings.SILENT)
            return part_in_lib, new_part
        if not os.path.isfile(template_path):
            cprint(f'[KCAD]\tError loading template file ({template_path})', silent=settings.SILENT)
            return part_in_lib, new_part

        # Load library
        schlib = SchLib(library_path)
        library_name = library_path.split(os.sep)[-1]
        cprint('[KCAD]\tNumber of parts in library ' + library_name + ': ' + str(schlib.getComponentCount()), silent=settings.SILENT)

        # Check if part already in library
        try:
            is_component_in_library = self.is_component_in_library(schlib, component_data['IPN'])
            part_in_lib = True
        except:
            is_component_in_library = False
        if is_component_in_library:
            return part_in_lib, new_part

        # Progress Update
        if show_progress and not progress.update_progress_bar_window():
            return part_in_lib, new_part

        # Load template
        templatelib = SchLib(template_path)
        # Load new component
        if templatelib.getComponentCount() == 1:
            for component in templatelib.components:
                new_component = component
        else:
            cprint('[KCAD]\tError: Found more than 1 component template in template file, aborting', silent=settings.SILENT)
            return part_in_lib, new_part

        # Update comment, name and definition
        new_component.comments[1] = '# ' + component_data['IPN'] + '\n'
        new_component.name = component_data['IPN']
        new_component.definition['name'] = component_data['IPN']

        # Update documentation
        new_component.documentation['description'] = component_data['description']
        new_component.documentation['datasheet'] = component_data['inventree_url']
        new_component.documentation['keywords'] = component_data['keywords']
        # cprint(new_component.documentation, silent=SILENT)

        # Update fields
        manufacturer_name = ''
        for field_idx in range(len(new_component.fields)):
            if 'name' in new_component.fields[field_idx]:
                component_field = str(new_component.fields[field_idx]['name']).replace('"', '')
                if component_field in component_data['parameters'].keys():
                    new_component.fields[field_idx]['name'] = component_data['parameters'][component_field]
                elif component_field == 'IPN':
                    new_component.fields[field_idx]['name'] = component_data['IPN']
                elif component_field == 'Manufacturer':
                    for manufacturer in component_data['manufacturer'].keys():
                        manufacturer_name = manufacturer
                        new_component.fields[field_idx]['name'] = manufacturer_name
                        break
                elif component_field == 'MPN':
                    new_component.fields[field_idx]['name'] = component_data['manufacturer'][manufacturer_name][0]

        schlib.addComponent(new_component)
        schlib.save()
        cprint(f'[KCAD]\tSuccess: Component added to library {library_name}', silent=settings.SILENT)
        part_in_lib = True
        new_part = True

        # Progress Update
        if show_progress and not progress.update_progress_bar_window():
            pass

        return part_in_lib, new_part

    def delete_component_from_lib(self, part_number, library_path):
        ''' Remove component (symbol) from KiCad library '''

        # Check that Library file exists
        if not os.path.isfile(library_path):
            cprint(f'[KCAD]\tError loading library file ({library_path})', silent=settings.SILENT)
            cprint('[KCAD]\tCheck library file path and name', silent=settings.SILENT)
            return False

        schlib = SchLib(library_path)
        library_name = library_path.split(os.sep)[-1]
        cprint('[KCAD]\tNumber of parts in library ' + library_name + ': ' + str(schlib.getComponentCount()), silent=settings.SILENT)

        try:
            schlib.removeComponent(part_number)
            schlib.save()
            cprint(f'[KCAD]\tSuccess: Component {part_number} was removed from library', silent=settings.SILENT)
            return True
        except:
            cprint(f'[KCAD]\tError: Component {part_number} was not found in library {library_name} (no delete)', silent=settings.SILENT)
            return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kintree',
 'kintree.common',
 'kintree.config',
 'kintree.database',
 'kintree.kicad',
 'kintree.search',
 'kintree.wrapt_timeout_decorator']

package_data = \
{'': ['*'],
 'kintree.config': ['digikey/*',
                    'inventree/*',
                    'kicad/*',
                    'lcsc/*',
                    'search/*',
                    'settings/*'],
 'kintree.kicad': ['templates/*', 'templates_project/*']}

install_requires = \
['PySimpleGUI>=4.28.0,<5.0',
 'PyYAML>=5.3.1,<6.0',
 'digikey-api>=0.4.0,<1.0',
 'fuzzywuzzy>=0.18.0,<1.0',
 'inventree>=0.2.4,<1.0',
 'multiprocess>=0.70.12.2,<0.71',
 'python-Levenshtein>=0.12.2,<0.13.0',
 'validators>=0.18.2',
 'wrapt>=1.12.1,<2.0']

entry_points = \
{'console_scripts': ['kintree = kintree:kintree_gui.main',
                     'kintree_setup_inventree = kintree:setup_inventree.main']}

setup_kwargs = {
    'name': 'kintree',
    'version': '0.4.1',
    'description': 'Fast part creation in KiCad and InvenTree',
    'long_description': '# Ki-nTree\n### Fast part creation in [KiCad](https://kicad-pcb.org/) and [InvenTree](https://inventree.readthedocs.io/) \n[![License: GPL v3.0](https://img.shields.io/badge/license-GPL_v3.0-green.svg)](https://www.gnu.org/licenses/gpl-3.0)\n[![Python Versions](https://raw.githubusercontent.com/sparkmicro/Ki-nTree/master/images/python_versions.svg)](https://www.python.org/)\n[![PyPI](https://img.shields.io/pypi/v/kintree)](https://pypi.org/project/kintree/)\n[![Tests | Linting | Publishing](https://github.com/sparkmicro/Ki-nTree/actions/workflows/test_deploy.yaml/badge.svg)](https://github.com/sparkmicro/Ki-nTree/actions)\n[![Coverage Status](https://coveralls.io/repos/github/sparkmicro/Ki-nTree/badge.svg?branch=master&service=github)](https://coveralls.io/github/sparkmicro/Ki-nTree?branch=master)\n\n## Demo Videos :fast_forward: [Full Demo](https://youtu.be/haSAu926BOI) :fast_forward: [KiCad Demo](https://youtu.be/NSMfCCD0uVw)\n\n<img src="https://raw.githubusercontent.com/sparkmicro/Ki-nTree/master/images/doc/kintree_example.png"  width="auto" height="auto">\n\n## Introduction\nKi-nTree (pronounced "Key Entry" or "Key \'n\' Tree") aims to:\n* automate part creation of KiCad library parts\n* automate part creation of InvenTree parts\n* synchronize parts data between KiCad and InvenTree\n\nKi-nTree works with:\n- Digi-Key\'s **enormous** part database and [free API](https://developer.digikey.com/)\n- the awesome open-source [Digi-Key API python library](https://github.com/peeter123/digikey-api) built and maintained by [@peeter123](https://github.com/peeter123)\n- the awesome open-source [InvenTree Inventory Management System](https://github.com/inventree/inventree) built and maintained by [@SchrodingersGat](https://github.com/SchrodingersGat)\n- [KiCad](https://kicad-pcb.org/) (of course!) and their open-source [library utils](https://github.com/KiCad/kicad-library-utils)\n\nKi-nTree was developped by [@eeintech](https://github.com/eeintech) for [SPARK Microsystems](https://www.sparkmicro.com/), who generously accepted to make it open-source!\n\n## Get Started\n\n### Requirements\n\n* Ki-nTree is currently tested for Python 3.7 to 3.9 versions.\n* Ki-nTree requires a Digi-Key **production** API instance. To create one, go to https://developer.digikey.com/. Create an account, an organization and add a **production** API to your organization. Save both Client ID and Secret keys.\n> [Here is a video](https://youtu.be/OI1EGEc0Ju0) to help with the different steps\n\n### Installation (system wide)\n\n1. Install using Pip\n\n``` bash\n$ pip install -U kintree\n```\n\n2. Run Ki-nTree\n\n``` bash\n$ kintree\n```\n\n### Run in virtual environment (contained)\n\n##### Linux / MacOS\n\nCreate a virtual environment and activate it with:\n\n``` bash\n$ python3 -m venv env-kintree\n$ source env-kintree/bin/activate\n```\n\nThen follow the steps from the [installation section](#installation-system-wide).\n\n##### Windows\n\nIn Git Bash, use the following commands to create and activate a virtual environment:\n``` bash\n$ python3 -m venv env-kintree\n$ source env-kintree/Scripts/activate\n```\nFor any other Windows terminal, refer to the [official documentation](https://docs.python.org/library/venv.html#creating-virtual-environments)\n\n### Usage Instructions\n\n#### Settings\n1. With Ki-nTree GUI open, click on "Settings > Digi-Key" and fill in both Digi-Key API Client ID and Secret keys (optional: click on "Test" to [get an API token](#get-digi-key-api-token))\n2. Click on "Settings > KiCad", browse to the location where KiCad symbol and footprint libraries are stored on your computer then click "Save"\n3. If you intend to use InvenTree with this tool, click on "Settings > InvenTree" and fill in your InvenTree server address and credentials then click "Save" (optional: click on "Test" to get an API token)\n\n#### Get Digi-Key API token\n<details>\n<summary>Show steps (click to expand)</summary>\n<p>\n\nEnter your Digi-Key developper account credentials then login. The following page will appear (`user@email.com` will show your email address):\n\n<img src="https://raw.githubusercontent.com/sparkmicro/Ki-nTree/master/images/doc/digikey_api_approval_request.png" width="600" height="auto">\n\nClick on "Allow", another page will open.  \nClick on the "Advanced" button, then click on "Proceed to localhost (unsafe)" at the bottom of the page:\n\n<img src="https://raw.githubusercontent.com/sparkmicro/Ki-nTree/master/images/doc/digikey_api_approval_request2.png"  width="600" height="auto">\n\n> On Chrome, if the "Proceed to localhost (unsafe)" link does not appear, enable the following flag: [chrome://flags/#allow-insecure-localhost](chrome://flags/#allow-insecure-localhost)\n\nLastly, a new page will open with a "You may now close this window." message, proceed to get the token.\n\n</p>\n</details>\n\n#### Part Number Search\n1. In the main window, enter the part number and click "OK", it will start by fetching part data using the Digi-Key API\n2. In the case the Digi-Key API token is not found or expired, a browser window will pop-up. To get a new token: [follow those steps](#get-digi-key-api-token)\n3. Once the part data has been successfully fetched from Digi-Key, you will be prompted to add/confirm/edit the `Category` and `Subcategory` to use for this part (Ki-nTree tries to match them automatically)  \n4. Then, you will be prompted with selecting the KiCad symbol library, template and footprint library to use for this part  \n5. It will take some time to complete the part creation in InvenTree and/or KiCad, once it finishes you\'ll be notified of the result  \n6. Finally, if the part was created or found in InvenTree, your browser will automatically open a new tab with the part information\n\nEnjoy!\n\n*For any problem/bug you find, please [report an issue](https://github.com/sparkmicro/Ki-nTree/issues).*\n\n## Development\n\n### Requirements\n\nYou need `python>=3.7` and `poetry`.\n\nYou can install poetry by following the instructions [on its official\nwebsite](https://python-poetry.org/docs/master/#installation), by using `pip\ninstall poetry` or by installing a package on your Linux distro.\n\n### Setup and run\n1. Clone this repository\n``` bash\n$ git clone https://github.com/sparkmicro/Ki-nTree\n```\n\n2. Install the requirements into a `poetry`-managed virtual environment\n``` bash\n$ poetry install\nInstalling dependencies from lock file\n...\nInstalling the current project: kintree (0.1.0)\n```\n> Note: the version is not accurate (placeholder only)\n\n3. Run Ki-nTree in the virtual environment\n```bash\n$ poetry run python -m kintree.kintree_gui\n```\nor\n\n```bash\n$ poetry shell\n$ python -m kintree.kintree_gui\n```\n\n#### Build\n1. Make sure you followed the previous installation steps, then run:\n``` bash\n$ poetry build\n```\n\nYou will get a message similar to this:\n\n```\nBuilding kintree (0.3.10)\n  - Building sdist\n  - Built kintree-0.3.10.tar.gz\n  - Building wheel\n  - Built kintree-0.3.10-py3-none-any.whl\n```\n2. Exit the virtual environment (`Ctrl + D` on Linux; you can also close the\n   terminal and reopen it in the same folder).\n\n   Run `pip install dist/<wheel_file>.whl` with the file name from the previous\n   step. For example:\n\n```bash\npip install dist/kintree-0.3.10-py3-none-any.whl\n```\n\n3. You can now start Ki-nTree by typing `kintree` in the terminal, provided\n   that your python dist path is a part of your `$PATH`.\n\n## Roadmap\n\n#### Versions 0.4.x or later\n##### Global\n\n- Support other suppliers like LCSC and Mouser ([Ref #50](https://github.com/sparkmicro/Ki-nTree/issues/50))\n- Allow user to decide the category code to use for IPN\n- Add "Synchronize" menu option to pull InvenTree parts data into KiCad\n- Fetch Digi-Key price breakdown and add it to Supplier Part (InvenTree)\n- Add option to add as alternate (supplier part) to existing part\n- Document configuration and backend\n\n##### GUI\n\n- Move to PySide2? ([Ref #37](https://github.com/sparkmicro/Ki-nTree/issues/37))\n- Add icon to GUI (not successful using PySimpleGUI)\n- Create loading animation for API searches (asyncio)\n',
    'author': 'eeintech',
    'author_email': 'eeintech@eeinte.ch',
    'maintainer': 'eeintech',
    'maintainer_email': 'eeintech@eeinte.ch',
    'url': 'https://github.com/sparkmicro/Ki-nTree',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
